/**
 * Entry point and game code.
 *
 * Copyright (C) 2006  Morten Hustveit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <SDL/SDL.h>

#ifndef WIN32
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#else
#include <winsock.h>
#include <io.h>
#endif

#include <errno.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>

#include "error.h"
#include "packet.h"
#include "sound.h"

#ifndef WIN32
typedef int SOCKET;
const int INVALID_SOCKET = -1;
const int SOCKET_ERROR = -1;

#define Sleep(n) usleep((n) * 1000)
#define send(a,b,c,d) write((a),(b),(c))
#define recv(a,b,c,d) read((a),(b),(c))
#endif

int hwalpha;

void load_images();
SDL_Surface* get_image(const char* name);

void load_font();
int has_char(int ch);
void print_string(int x, int y, const wchar_t* string);

const int width = 640;
const int height = 480;
int fullscreen = 1;
SDL_Surface* screen;

unsigned short listenport;
SOCKET listenfd;
SOCKET serverfd;
SOCKET outpeerfd;
SOCKET inpeerfd;
int outpeerconnected;
int outpeervalid;
int inpeervalid;
char handshake[4];
static char inbufferip[sizeof(struct data_packet)];
static char inbufferop[sizeof(struct data_packet)];
static char inbuffer[sizeof(struct data_packet)];
static char outbufferip[sizeof(struct data_packet) * 32];
static char outbufferop[sizeof(struct data_packet) * 32];
static char outbuffer[sizeof(struct data_packet) * 32];
static int inbufferip_state;
static int inbufferop_state;
static int inbuffer_state;
static int outbufferip_state;
static int outbufferop_state;
static int outbuffer_state;

Uint32 last_tick = 0;

SDL_Surface* splash;
SDL_Surface* logo;
SDL_Surface* background;
SDL_Surface* base[129];
SDL_Surface* bubbles[8];

const int time_stepms = 8;
const float time_step = 0.008f;
const int field_height = 14;
const int max_field_width = 8;
const float bubble_speed = 500;
const int max_x = 414 - 190;
#define WIDTH(y) (((y) & 1) ? 7 : 8)

enum mode
{
  MODE_MAIN_MENU,
  MODE_CREATE_GAME,
  MODE_WAITING,
  MODE_JOIN_GAME,
  MODE_GAME,
  MODE_SCORE,
  MODE_ABORT_MESSAGE
};

enum mode mode = MODE_MAIN_MENU;
wchar_t game_name[33];
wchar_t message[256];
int won;
int is_server;
int input_locked;
int single_player = 0;

enum flag
{
  FLAG_KEY_LEFT = 0x01,
  FLAG_KEY_RIGHT = 0x02,
  FLAG_BUBBLES = 0x04,
  FLAG_VICTORY = 0x08,
  FLAG_LOSS = 0x10,
  FLAG_READY = 0x20
};

struct moving_bubble
{
  float x;
  float y;
  float velx;
  float vely;

  int lastpaintx;
  int lastpainty;

  unsigned char falling;
  unsigned char color;
};

struct player_state
{
  unsigned char field[14][8];

  int bubble;
  int next_bubble;

  float angle;
  int right;

  int dirty_minx;
  int dirty_miny;
  int dirty_maxx;
  int dirty_maxy;
  float last_angle;

  int score;
  int ready;

  struct moving_bubble mbubbles[128];

  unsigned char evil_bubbles[128];
  int evil_bubble_count;
  unsigned int evil_bubble_seed;
};

static struct player_state players[2];

static void init_field();

#define SET_RECT(r, xv, yv, wv, hv) do { (r).x = (xv); (r).y = (yv); (r).w = (wv); (r).h = (hv); } while(0)

static unsigned int rng_seed = 0;

static unsigned int rng()
{
  rng_seed = (rng_seed + 0xA853753) % 177
           + (rng_seed * 0x8034052);

  return rng_seed;
}

static void pack_float(unsigned char* a, float v)
{
  union
  {
    int ival;
    float fval;
  } u;

  u.fval = v;

  a[0] = u.ival >> 24;
  a[1] = u.ival >> 16;
  a[2] = u.ival >> 8;
  a[3] = u.ival;
}

static float unpack_float(const unsigned char* a)
{
  union
  {
    int ival;
    float fval;
  } u;

  u.ival = (a[0] << 24) | (a[1] << 16) | (a[2] << 8) | a[3];

  return u.fval;
}

static void mark_dirty(struct player_state* p, int x, int y, int width, int height)
{
  if(x < p->dirty_minx)
    p->dirty_minx = x;

  if(y < p->dirty_miny)
    p->dirty_miny = y;

  if(x + width > p->dirty_maxx)
    p->dirty_maxx = x + width;

  if(y + height > p->dirty_maxy)
    p->dirty_maxy = y + height;
}

static void scan(struct player_state* p, int x, int y, int color, int* matches, int* match_count)
{
  int i, idx;

  if(x < 0 || x >= WIDTH(y) || y < 0 || y > field_height)
    return;

  if((color && (p->field[y][x] != color))
  || (!color && !p->field[y][x]))
    return;

  idx = y * max_field_width + x;

  for(i = 0; i < *match_count; ++i)
    if(matches[i] == idx)
      return;

  matches[(*match_count)++] = idx;

  scan(p, x - 1, y, color, matches, match_count);
  scan(p, x + 1, y, color, matches, match_count);

  if(y & 1)
  {
    scan(p, x, y - 1, color, matches, match_count);
    scan(p, x + 1, y - 1, color, matches, match_count);
    scan(p, x, y + 1, color, matches, match_count);
    scan(p, x + 1, y + 1, color, matches, match_count);
  }
  else
  {
    scan(p, x - 1, y - 1, color, matches, match_count);
    scan(p, x, y - 1, color, matches, match_count);
    scan(p, x - 1, y + 1, color, matches, match_count);
    scan(p, x, y + 1, color, matches, match_count);
  }
}

static void remove_bubble(struct player_state* p, int x, int y)
{
  int i;
  struct moving_bubble mbubble;

  mbubble.falling = 1;
  mbubble.x = x * 32 + ((y & 1) ? 16 : 0);
  mbubble.y = y * 28;
  mbubble.velx = sinf(powf(x, 1.4f) + powf(y, 2.8f)) * 0.5f * 200.0f;
  mbubble.vely = (sinf(powf(x, 3.4f) + powf(y, 0.6f)) - 1.0f) * 25.0f - 300.0f;
  mbubble.color = p->field[y][x];
  mbubble.lastpaintx = mbubble.x;
  mbubble.lastpainty = mbubble.y;

  for(i = 0; i < sizeof(p->mbubbles) / sizeof(p->mbubbles[0]); ++i)
  {
    if(p->mbubbles[i].color == 0 && p->mbubbles[i].lastpaintx == INT_MIN)
    {
      p->mbubbles[i] = mbubble;

      break;
    }
  }

  p->field[y][x] = 0;
}

/* XXX: Modifies p->payload_size */
static void send_packet(struct data_packet* p)
{
  int packet_size = 3 + p->payload_size;
  int* state = 0;
  char* buffer;
  SOCKET* fd;

  if(single_player)
    return;

  if(p->packet_type >= 0x10)
  {
    if(inpeerfd != INVALID_SOCKET && inpeervalid)
      state = &outbufferip_state, buffer = outbufferip, fd = &inpeerfd;
    else if(outpeerfd != INVALID_SOCKET && outpeervalid)
      state = &outbufferop_state, buffer = outbufferop, fd = &outpeerfd;
  }

  if(!state)
    state = &outbuffer_state, buffer = outbuffer, fd = &serverfd;

  if(*state + packet_size > sizeof(outbuffer))
  {
    close(*fd);
    *fd = INVALID_SOCKET;

    wcscpy(message, L"Network Output Buffer Full");

    mode = MODE_ABORT_MESSAGE;

    return;
  }

  p->payload_size = htons(p->payload_size);

  memcpy(buffer + *state, p, packet_size);
  *state += packet_size;
}

static void init_movement_packet(struct data_packet* p)
{
  p->packet_type = packet_movement;
  p->payload_size = 6;

  pack_float(p->movement.angle, players[0].angle);
  p->movement.flags = 0;

  if(players[0].right == 1)
    p->movement.flags |= FLAG_KEY_RIGHT;
  else if(players[0].right == -1)
    p->movement.flags |= FLAG_KEY_LEFT;

  p->movement.bubbles = players[0].bubble | (players[0].next_bubble << 4);
}

static void stick(struct player_state* p, int bx, int by, int color)
{
  int i, j, k;
  int matches[max_field_width * field_height];
  int match_count = 0;

  if(p == &players[0])
    sounds[3].pos = 0;

  p->field[by][bx] = color;
  mark_dirty(p, bx * 32 + ((by & 1) ? 16 : 0), by * 28, 32, 32);

  scan(p, bx, by, color, matches, &match_count);

  if(match_count >= 3)
  {
    if(p == &players[0])
      sounds[0].pos = 0;

    for(i = 0; i < match_count; ++i)
    {
      remove_bubble(p, matches[i] % max_field_width,
                    matches[i] / max_field_width);
    }

    match_count = 0;

    for(i = 0; i < WIDTH(0); ++i)
      scan(p, i, 0, 0, matches, &match_count);

    if(!match_count)
    {
      if(single_player)
      {
        init_field();

        last_tick = SDL_GetTicks();
      }
      else if(is_server)
      {
        struct data_packet packet;

        init_movement_packet(&packet);

        if(p == &players[0])
        {
          packet.movement.flags |= FLAG_VICTORY;

          won = 1;
          ++players[0].score;
        }
        else
        {
          packet.movement.flags |= FLAG_LOSS;

          won = 0;
          ++players[1].score;
        }

        send_packet(&packet);

        players[0].ready = 0;
        players[1].ready = 0;

        mode = MODE_SCORE;
      }
      else
        input_locked = 1;
    }
    else
    {
      for(i = 0; i < field_height; ++i)
      {
        for(j = 0; j < WIDTH(i); ++j)
        {
          if(!p->field[i][j])
            continue;

          for(k = 0; k < match_count; ++k)
            if(matches[k] == i * max_field_width + j)
              break;

          if(k == match_count)
          {
            if(p == &players[0])
            {
              k = players[1].evil_bubble_count++;

              players[1].evil_bubbles[k] = p->field[i][j];
            }
            else
            {
              k = players[0].evil_bubble_count++;

              players[0].evil_bubbles[k] = p->field[i][j];
            }

            remove_bubble(p, j, i);
          }
        }
      }
    }
  }
}

static void cond_blit(struct player_state* p, SDL_Surface* source, SDL_Rect* source_rect, SDL_Surface* dest, SDL_Rect* dest_rect)
{
  int xoff, yoff;

  if(p == &players[0])
    xoff = 30, yoff = 40;
  else
    xoff = 355, yoff = 40;

  if(p->dirty_minx >= p->dirty_maxx)
    return;

  if(!dest_rect
  || (   dest_rect->x <= p->dirty_maxx + xoff
      && dest_rect->x + dest_rect->w >= p->dirty_minx + xoff
      && dest_rect->y <= p->dirty_maxy + yoff
      && dest_rect->y + dest_rect->h >= p->dirty_miny + yoff))
  {
    SDL_BlitSurface(source, source_rect, dest, dest_rect);
  }
}

static void game_tick(int paint)
{
  int i, j, k;
  int player;
  int yoff, xoff;
  int bx, by;
  SDL_Rect rect;
  SDL_Rect rect2;

  for(player = 0; player < 2; ++player)
  {
    struct player_state* p = &players[player];

    if(single_player && player > 0)
      break;

    if(player == 0)
      xoff = 30, yoff = 40;
    else
      xoff = 355, yoff = 40;

    p->angle += p->right * 90 * time_step;

    if(p->angle < -85)
      p->angle = -85;
    else if(p->angle > 85)
      p->angle = 85;

    for(k = 0; k < sizeof(p->mbubbles) / sizeof(p->mbubbles[0]); ++k)
    {
      struct moving_bubble* b = &p->mbubbles[k];

      if(!b->color)
        continue;

      if(!b->falling)
      {
        b->x += b->velx * time_step;
        b->y += b->vely * time_step;

        if(b->x < 0)
        {
          b->x = -b->x;
          b->velx = -b->velx;

          if(player == 0)
            sounds[2].pos = 0;
        }
        else if(b->x > max_x)
        {
          b->x = 2 * max_x - b->x;
          b->velx = -b->velx;

          if(player == 0)
            sounds[2].pos = 0;
        }

        for(i = -1; i < field_height; ++i)
        {
          for(j = 0; j < WIDTH(i); ++j)
          {
            float posx, posy;
            float dirx, diry;

            if(i != -1 && !p->field[i][j])
              continue;

            posx = j * 32.0f + ((i & 1) ? 32.0f : 16.0f);
            posy = i * 28.0f + 16.0f;

            dirx = posx - (b->x + 16.0f);
            diry = posy - (b->y + 16.0f);

            if(dirx * dirx + diry * diry < 784.0f)
            {
              if(fabs(dirx) > fabs(diry))
              {
                by = i;

                if(dirx > 0)
                  bx = j - 1;
                else
                  bx = j + 1;
              }
              else // fabs(dirx) <= fabs(diry)
              {
                if(diry > 0)
                  by = i - 1;
                else
                  by = i + 1;

                if(by & 1)
                {
                  if(dirx > 0)
                    bx = j - 1;
                  else
                    bx = j;
                }
                else
                {
                  if(dirx > 0)
                    bx = j;
                  else
                    bx = j + 1;
                }
              }

              if(by > 11)
              {
                if(single_player)
                {
                  SDL_BlitSurface(background, 0, screen, 0);

                  init_field();

                  last_tick = SDL_GetTicks();
                  game_tick(1);
                }
                else if(is_server)
                {
                  struct data_packet packet;

                  init_movement_packet(&packet);

                  if(p == &players[0])
                  {
                    packet.movement.flags |= FLAG_LOSS;

                    won = 0;
                    ++players[1].score;
                  }
                  else
                  {
                    packet.movement.flags |= FLAG_VICTORY;

                    won = 1;
                    ++players[0].score;
                  }

                  send_packet(&packet);

                  players[0].ready = 0;
                  players[1].ready = 0;

                  mode = MODE_SCORE;
                }
                else
                  input_locked = 1;
              }

              stick(p, bx, by, b->color);

              b->color = 0;

              goto collide;
            }
          }
        }

collide:;

        if(mode == MODE_SCORE)
          return;
      }
      else
      {
        b->vely += 1500.0f * time_step;
        b->x += b->velx * time_step;
        b->y += b->vely * time_step;

        if(b->y > 480)
          b->color = 0;
      }
    }

    if(paint)
    {
      if(p->dirty_minx < p->dirty_maxx)
      {
        SET_RECT(rect, xoff + p->dirty_minx, yoff + p->dirty_miny,
                 p->dirty_maxx - p->dirty_minx,
                 p->dirty_maxy - p->dirty_miny);
        SDL_BlitSurface(background, &rect, screen, &rect);
      }

      for(i = 0; i < sizeof(p->mbubbles) / sizeof(p->mbubbles[0]); ++i)
      {
        if(p->mbubbles[i].lastpaintx == INT_MIN)
          continue;

        SET_RECT(rect, p->mbubbles[i].lastpaintx + xoff, p->mbubbles[i].lastpainty + yoff, 32, 32);

        SDL_BlitSurface(background, &rect, screen, &rect);

        mark_dirty(p, p->mbubbles[i].lastpaintx, p->mbubbles[i].lastpainty, 32, 32);

        p->mbubbles[i].lastpaintx = INT_MIN;
      }

      if(p->dirty_minx < p->dirty_maxx)
      {
        for(i = 0; i < 10; ++i)
        {
          for(j = 0; j < WIDTH(i); ++j)
          {
            if(!p->field[i][j])
              continue;

            rect.x = xoff + j * 32 + ((i & 1) ? 16 : 0);
            rect.y = yoff + i * 28;
            rect.w = 32;
            rect.h = 32;

            cond_blit(p, bubbles[p->field[i][j] - 1], 0, screen, &rect);
          }
        }
      }

      if(p->last_angle != p->angle
      || (   78 <= p->dirty_maxx && 178 >= p->dirty_minx
          && 316 <= p->dirty_maxy && 432 >= p->dirty_miny))
      {
        mark_dirty(p, 78, 316, 100, 100);

        p->last_angle = p->angle;
      }

      if(p->dirty_minx < p->dirty_maxx)
      {
        rect.x = xoff + 78;
        rect.y = yoff + 316;
        rect.w = 100;
        rect.h = 100;

        cond_blit(p, background, &rect, screen, &rect);

        for(i = 10; i < field_height; ++i)
        {
          for(j = 0; j < WIDTH(i); ++j)
          {
            if(!p->field[i][j])
              continue;

            rect2.x = xoff + j * 32 + ((i & 1) ? 16 : 0);
            rect2.y = yoff + i * 28;
            rect2.w = 32;
            rect2.h = 32;

            cond_blit(p, bubbles[p->field[i][j] - 1], 0, screen, &rect2);
          }
        }

        if(78 <= p->dirty_maxx && 178 >= p->dirty_minx
        && 316 <= p->dirty_maxy && 432 >= p->dirty_miny)
        {
          rect2.x = xoff + 112;
          rect2.y = yoff + 350;
          rect2.w = 32;
          rect2.h = 32;

          SDL_BlitSurface(bubbles[p->bubble], 0, screen, &rect2);
          SDL_BlitSurface(base[(int) ((p->angle + 90.0f) * 128 / 180.0f)], 0, screen, &rect);

          rect2.x = xoff + 112;
          rect2.y = yoff + 400;

          SDL_BlitSurface(bubbles[p->next_bubble], 0, screen, &rect2);
        }
      }

      p->dirty_minx = max_field_width * 32;
      p->dirty_miny = 440;
      p->dirty_maxx = 0;
      p->dirty_maxy = 0;

      for(i = 0; i < sizeof(p->mbubbles) / sizeof(p->mbubbles[0]); ++i)
      {
        if(!p->mbubbles[i].color)
          continue;

        SET_RECT(rect,
                 (int) p->mbubbles[i].x + xoff,
                 (int) p->mbubbles[i].y + yoff,
                 32, 32);

        SDL_BlitSurface(bubbles[p->mbubbles[i].color - 1], 0, screen, &rect);

        p->mbubbles[i].lastpaintx = rect.x - xoff;
        p->mbubbles[i].lastpainty = rect.y - yoff;
      }
    }
  }

  if(single_player)
  {
    print_string(480, 150, L"Waiting for");
    print_string(480, 190, L"Other Player");
  }
}

static int random_bubble(struct player_state* p)
{
  int i, j, result;

  for(;;)
  {
    result = rand() % 8;

    for(i = 0; i < field_height; ++i)
    {
      for(j = 0; j < WIDTH(i); ++j)
      {
        if(p->field[i][j] == result + 1)
          return result;
      }
    }
  }
}

static void shoot(struct player_state* p)
{
  int i;
  struct moving_bubble mbubble;

  mbubble.falling = 0;
  mbubble.velx = sin(p->angle / 180.0f * M_PI) * bubble_speed;
  mbubble.vely = -cos(p->angle / 180.0f * M_PI) * bubble_speed;
  mbubble.x = 112.0f;
  mbubble.y = 350.0f;
  mbubble.color = p->bubble + 1;

  for(i = 0; i < sizeof(p->mbubbles) / sizeof(p->mbubbles[0]); ++i)
  {
    if(p->mbubbles[i].color == 0)
    {
      mbubble.lastpaintx = p->mbubbles[i].lastpaintx;
      mbubble.lastpainty = p->mbubbles[i].lastpainty;
      p->mbubbles[i] = mbubble;

      break;
    }
  }

  p->bubble = p->next_bubble;

  if(p != &players[1])
    p->next_bubble = random_bubble(p);

  sounds[1].pos = 0;

  while(p->evil_bubble_count && i < sizeof(p->mbubbles) / sizeof(p->mbubbles[0]))
  {
    if(p->mbubbles[i].color != 0 || p->mbubbles[i].lastpaintx != INT_MIN)
    {
      ++i;

      continue;
    }

    float rand = sin(pow(p->evil_bubble_seed++, 4.5)) * 0.5 + 0.5;
    float angle = rand * 60 - 30;

    p->mbubbles[i].falling = 0;
    p->mbubbles[i].velx = sin(angle / 180 * M_PI) * bubble_speed;
    p->mbubbles[i].vely = -cos(angle / 180 * M_PI) * bubble_speed;
    p->mbubbles[i].x = 112.0f;
    p->mbubbles[i].y = 400.0f;
    p->mbubbles[i].color = p->evil_bubbles[0];

    --p->evil_bubble_count;

    memmove(p->evil_bubbles, &p->evil_bubbles[1], p->evil_bubble_count);

    ++i;
  }
}

static void init_player(struct player_state* p)
{
  int i;

  mark_dirty(p, 0, 0, max_field_width * 32, 440);
  p->last_angle = -1.0f;
  p->angle = 0.0f;
  p->right = 0;
  p->evil_bubble_count = 0;
  p->evil_bubble_seed = 0;

  for(i = 0; i < sizeof(p->mbubbles) / sizeof(p->mbubbles[0]); ++i)
  {
    p->mbubbles[i].color = 0;
    p->mbubbles[i].lastpaintx = INT_MIN;
  }
}

static void init_field()
{
  int i, j, color;

  for(i = 0; i < field_height / 2; ++i)
  {
    for(j = 0; j < WIDTH(i); ++j)
    {
      color = (rng() % 8) + 1;

      players[0].field[i][j] = color;
      players[1].field[i][j] = color;
    }
  }

  for(; i < field_height; ++i)
  {
    for(j = 0; j < WIDTH(i); ++j)
    {
      players[0].field[i][j] = 0;
      players[1].field[i][j] = 0;
    }
  }

  players[0].bubble = rng() % 8;
  players[0].next_bubble = rng() % 8;
  players[1].bubble = players[0].bubble;
  players[1].next_bubble = players[0].next_bubble;

  init_player(&players[0]);
  init_player(&players[1]);

  input_locked = 0;
}

void connect_to_master()
{
  int tcp_port = 7170;
  struct hostent* host;
  struct sockaddr_in address;
  struct data_packet packet;

  serverfd = INVALID_SOCKET;
  outbuffer_state = 0;
  inbuffer_state = 0;

  host = gethostbyname("junoplay.com");

  if(!host)
    return;

  if(INVALID_SOCKET == (serverfd = socket(PF_INET, SOCK_STREAM, 0)))
    return;

  address.sin_family = AF_INET;
  address.sin_port = htons(tcp_port);
  memcpy(&address.sin_addr, host->h_addr, host->h_length);

  if(SOCKET_ERROR == connect(serverfd, (struct sockaddr*) &address, sizeof(address)))
  {
    close(serverfd);
    serverfd = INVALID_SOCKET;

    return;
  }

  packet.packet_type = packet_identify;
  packet.payload_size = sizeof(packet.identify);

  strcpy(packet.identify.game, "pengupop2006052100");

  if(listenfd == INVALID_SOCKET)
  {
    packet.identify.port_lo = 0xff;
    packet.identify.port_hi = 0xff;
  }
  else
  {
    packet.identify.port_lo = listenport;
    packet.identify.port_hi = listenport >> 8;
  }

  send_packet(&packet);
}

void start_listening()
{
  if(-1 == (listenfd = socket(PF_INET, SOCK_STREAM, 0)))
    return;

  for(listenport = 7171; listenport < 7180; ++listenport)
  {
    struct sockaddr_in address;

    address.sin_family = AF_INET;
    address.sin_addr.s_addr = INADDR_ANY;
    address.sin_port = htons(listenport);

    if(SOCKET_ERROR == bind(listenfd, (struct sockaddr*) &address,
                            sizeof(address)))
      continue;

    if(SOCKET_ERROR == listen(listenfd, 1))
      break;

    return;
  }

  close(listenfd);
  listenfd = INVALID_SOCKET;
}

void process_packet(struct data_packet* packet)
{
  int i, j, k, l;
  struct sockaddr_in address;

  if(packet->packet_type == packet_start_game)
  {
    if((mode == MODE_GAME && !single_player) || mode == MODE_SCORE)
    {
      close(serverfd);
      serverfd = INVALID_SOCKET;

      wcscpy(message, L"Corrupt network data");

      mode = MODE_ABORT_MESSAGE;

      return;
    }

    single_player = 0;

    memcpy(handshake, packet->start_game.seed, 4);
    rng_seed = packet->start_game.seed[0] << 24
             | packet->start_game.seed[1] << 16
             | packet->start_game.seed[2] << 8
             | packet->start_game.seed[3];

    SDL_BlitSurface(background, 0, screen, 0);

    is_server = packet->start_game.is_server;

    mode = MODE_GAME;

    init_field();
    players[0].score = 0;
    players[1].score = 0;
    players[0].ready = 1;
    players[1].ready = 1;

    last_tick = SDL_GetTicks();
    game_tick(1);

    outbufferop_state = 0;
    inbufferop_state = 0;
    outbufferip_state = 0;
    inbufferip_state = 0;
    outpeervalid = 0;
    inpeervalid = 0;

    if(packet->start_game.port_lo == 0xff
    && packet->start_game.port_hi == 0xff)
      return;

    if(INVALID_SOCKET == (outpeerfd = socket(PF_INET, SOCK_STREAM, 0)))
      return;

    /* We are connecting asynchronously to avoid deadlock */
    outpeerconnected = 0;

#ifndef WIN32
    if(-1 == fcntl(outpeerfd, F_SETFL, O_NONBLOCK))
    {
      close(outpeerfd);
      outpeerfd = INVALID_SOCKET;

      return;
    }
#else
    u_long one = 1;

    if(SOCKET_ERROR == ioctlsocket(outpeerfd, FIONBIO, &one))
    {
      close(outpeerfd);
      outpeerfd = INVALID_SOCKET;
    }
#endif

    address.sin_family = AF_INET;
    address.sin_port = htons(packet->start_game.port_lo | packet->start_game.port_hi << 8);
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
    address.sin_addr.s_addr
      = packet->start_game.other_host[0]
      | packet->start_game.other_host[1] << 8
      | packet->start_game.other_host[2] << 16
      | packet->start_game.other_host[3] << 24;
#else
    address.sin_addr.s_addr
      = packet->start_game.other_host[0] << 24
      | packet->start_game.other_host[1] << 16
      | packet->start_game.other_host[2] << 8
      | packet->start_game.other_host[3];
#endif

    if(SOCKET_ERROR == connect(outpeerfd, (struct sockaddr*) &address, sizeof(address)))
    {
#ifndef WIN32
      if(errno != EINPROGRESS)
#else
      if(errno != WSAEWOULDBLOCK)
#endif
      {
        close(outpeerfd);
        outpeerfd = INVALID_SOCKET;
      }
    }
  }
  else if(packet->packet_type == packet_movement)
  {
    if(packet->movement.flags & FLAG_READY)
    {
      players[1].ready = 1;

      if(players[0].ready)
      {
        SDL_BlitSurface(background, 0, screen, 0);

        mode = MODE_GAME;

        last_tick = SDL_GetTicks();
        game_tick(1);
      }
    }

    if(mode != MODE_GAME)
      return;

    players[1].angle = unpack_float(packet->movement.angle);

    if(packet->movement.flags & FLAG_KEY_RIGHT)
      players[1].right = 1;
    else if(packet->movement.flags & FLAG_KEY_LEFT)
      players[1].right = -1;
    else
      players[1].right = 0;

    if(packet->movement.flags & FLAG_BUBBLES)
    {
      k = 0;

      for(i = 0; i < field_height; ++i)
      {
        for(j = 0; j < WIDTH(i); ++j, ++k)
        {
          l = (packet->movement.field[k / 2] >> ((k & 1) ? 4 : 0)) & 0xF;

          if(players[1].field[i][j] != l)
          {
            mark_dirty(&players[1], j * 32 + ((i & 1) ? 16 : 0), i * 28, 32, 32);
            players[1].field[i][j] = l;
          }
        }
      }

      for(i = 0; i < sizeof(players[1].mbubbles) / sizeof(players[1].mbubbles[0]); ++i)
      {
        if(i < packet->movement.mbubble_count)
        {
          players[1].mbubbles[i].x = unpack_float(packet->movement.mbubbles[i].x);
          players[1].mbubbles[i].y = unpack_float(packet->movement.mbubbles[i].y);
          players[1].mbubbles[i].velx = unpack_float(packet->movement.mbubbles[i].velx);
          players[1].mbubbles[i].vely = unpack_float(packet->movement.mbubbles[i].vely);
          players[1].mbubbles[i].falling = packet->movement.mbubbles[i].falling;
          players[1].mbubbles[i].color = packet->movement.mbubbles[i].color;
        }
        else
        {
          players[1].mbubbles[i].color = 0;
        }
      }
    }

    if(packet->movement.flags & FLAG_VICTORY)
    {
      won = 0;
      ++players[1].score;

      players[0].ready = 0;
      players[1].ready = 0;

      mode = MODE_SCORE;

      return;
    }

    if(packet->movement.flags & FLAG_LOSS)
    {
      won = 1;
      ++players[0].score;

      players[0].ready = 0;
      players[1].ready = 0;

      mode = MODE_SCORE;

      return;
    }

    players[1].bubble = packet->movement.bubbles & 7;
    players[1].next_bubble = (packet->movement.bubbles >> 4) & 7;
  }
  else if(packet->packet_type == packet_abort_game)
  {
    int i;

    for(i = 0; i < 32; ++i)
      message[i] = packet->abort_game.reason[i];
    message[32] = 0;

    mode = MODE_ABORT_MESSAGE;
  }
  else
  {
    fprintf(stderr, "Received packet of type %u\n", packet->packet_type);
  }
}

void show_splash()
{
  SDL_Event event;

  SDL_BlitSurface(splash, 0, screen, 0);
  SDL_UpdateRect(screen, 0, 0, 0, 0);

  last_tick = SDL_GetTicks();

  while(SDL_GetTicks() - last_tick < 5000)
  {
    while(SDL_PollEvent(&event))
    {
      if(event.type == SDL_KEYDOWN)
      {
        if(event.key.keysym.sym == SDLK_ESCAPE
        || event.key.keysym.sym == SDLK_SPACE
        || event.key.keysym.sym == SDLK_RETURN)
          return;
      }
    }

    SDL_UpdateRect(screen, 0, 0, 0, 0);

    Sleep(10);
  }
}

#ifndef WIN32
int main(int argc, char** argv)
#else
int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
                   LPSTR cmdline, int cmdshow)
#endif
{
  int i, j, k, selection = 0;
  SDL_Event event;
  SDL_Rect rect;

#ifdef WIN32
  FreeConsole();

  WSADATA wsadata;

  WSAStartup(0x0101, &wsadata);
#else
  signal(SIGPIPE, SIG_IGN);
  signal(SIGALRM, SIG_IGN);
#endif

  if(-1 == SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO))
    fatal_error("SDL initialization failed: %s", SDL_GetError());

  atexit(SDL_Quit);

  SDL_EnableUNICODE(1);

  screen = SDL_SetVideoMode(width, height, 0, SDL_SWSURFACE | SDL_FULLSCREEN);

  if(hwalpha)
    info("Has HW alpha");

  if(!screen)
    fatal_error("Failed to create window: %s", SDL_GetError());

  SDL_WM_SetCaption("Pengupop", 0);
  SDL_ShowCursor(0);

  load_images();
  load_font();

  splash = get_image("splash.png");
  SDL_SetAlpha(splash, 0, 0);

  logo = get_image("logo.png");
  SDL_SetAlpha(logo, 0, 0);

  SDL_BlitSurface(logo, 0, screen, 0);
  SDL_UpdateRect(screen, 0, 0, 0, 0);

  background = get_image("backgrnd.png");
  SDL_SetAlpha(background, 0, 0);

  for(i = 0; i < 129; ++i)
  {
    char name[64];

    sprintf(name, "base.png$%d", i);

    base[i] = get_image(name);
  }

  for(i = 0; i < 8; ++i)
  {
    char name[64];

    sprintf(name, "bubble-colourblind-%d.png", i + 1);

    bubbles[i] = get_image(name);
  }

  outpeerfd = INVALID_SOCKET;
  inpeerfd = INVALID_SOCKET;

  start_listening();

  connect_to_master();

  sdl_audio.freq = 44100;
  sdl_audio.format = AUDIO_S16SYS;
  sdl_audio.channels = 1;
  sdl_audio.samples = 1024;
  sdl_audio.callback = sound_callback;
  sdl_audio.userdata = 0;

  if(-1 != SDL_OpenAudio(&sdl_audio, 0))
  {
    load_sounds();

    SDL_PauseAudio(0);
  }
  else
    sdl_audio.freq = 0;

  for(;;)
  {
    int update_movement = 0;
    int has_shot = 0;
    int result;

    if(mode != MODE_GAME && mode != MODE_SCORE)
    {
      if(inpeerfd != INVALID_SOCKET)
      {
        close(inpeerfd);
        inpeerfd = INVALID_SOCKET;
      }

      if(outpeerfd != INVALID_SOCKET)
      {
        close(outpeerfd);
        outpeerfd = INVALID_SOCKET;
      }
    }

    if(serverfd != INVALID_SOCKET)
    {
      fd_set readset, writeset;
      struct timeval timeout;
#ifndef WIN32
      int maxfd = 0;
#endif

      FD_ZERO(&readset);
      FD_ZERO(&writeset);

      FD_SET(serverfd, &readset);

      if(outbuffer_state)
        FD_SET(serverfd, &writeset);

#ifndef WIN32
      maxfd = serverfd;
#endif

      if(listenfd != INVALID_SOCKET)
      {
        FD_SET(listenfd, &readset);

#ifndef WIN32
        if(listenfd > maxfd)
          maxfd = listenfd;
#endif
      }

      if(outpeerfd != INVALID_SOCKET && mode != MODE_WAITING)
      {
        if(!outpeerconnected)
          FD_SET(outpeerfd, &writeset);
        else
        {
          if(outbufferop_state)
            FD_SET(outpeerfd, &writeset);

          FD_SET(outpeerfd, &readset);
        }

#ifndef WIN32
        if(outpeerfd > maxfd)
          maxfd = outpeerfd;
#endif
      }

      if(inpeerfd != INVALID_SOCKET && mode != MODE_WAITING)
      {
        if(outbufferip_state)
          FD_SET(inpeerfd, &writeset);

        FD_SET(inpeerfd, &readset);

#ifndef WIN32
        if(inpeerfd > maxfd)
          maxfd = inpeerfd;
#endif
      }

      timeout.tv_sec = 0;
      timeout.tv_usec = 0;

#ifndef WIN32
      int count = select(maxfd + 1, &readset, &writeset, 0, &timeout);
#else
      int count = select(0, &readset, &writeset, 0, &timeout);
#endif

      if(count == -1 && errno == EINTR)
        continue;

      if(count == -1)
      {
        close(serverfd);
        serverfd = INVALID_SOCKET;

        wcscpy(message, L"Connection Aborted (select)");

        mode = MODE_ABORT_MESSAGE;

        continue;
      }

      if(FD_ISSET(serverfd, &readset))
      {
        struct data_packet* packet = (struct data_packet*) inbuffer;

        result = recv(serverfd, inbuffer + inbuffer_state, sizeof(inbuffer) - inbuffer_state, 0);

        if(result <= 0)
        {
          close(serverfd);
          serverfd = INVALID_SOCKET;

          wcscpy(message, L"Connection Aborted (read)");

          mode = MODE_ABORT_MESSAGE;

          continue;
        }

        inbuffer_state += result;

        while(inbuffer_state >= 3 && inbuffer_state >= 3 + ntohs(packet->payload_size))
        {
          packet->payload_size = ntohs(packet->payload_size);
          process_packet(packet);

          inbuffer_state -= 3 + packet->payload_size;

          memmove(inbuffer, &inbuffer[3 + packet->payload_size],
                  inbuffer_state);
        }
      }

      if(FD_ISSET(serverfd, &writeset))
      {
        result = send(serverfd, outbuffer, outbuffer_state, 0);

        if(result <= 0)
        {
          close(serverfd);
          serverfd = INVALID_SOCKET;

          wcscpy(message, L"Connection Aborted (write)");

          mode = MODE_ABORT_MESSAGE;

          continue;
        }

        outbuffer_state -= result;
        memmove(outbuffer, outbuffer + result, outbuffer_state);
      }

      if(listenfd != INVALID_SOCKET && FD_ISSET(listenfd, &readset))
      {
        SOCKET newsocket = accept(listenfd, 0, 0);

        if(newsocket != INVALID_SOCKET)
        {
          if(inpeerfd != INVALID_SOCKET)
            close(newsocket);
          else
          {
            inpeerfd = newsocket;

            memcpy(outbufferip, handshake, sizeof(handshake));
            outbufferip_state = sizeof(handshake);
          }
        }
      }

      if(outpeerfd != INVALID_SOCKET && FD_ISSET(outpeerfd, &writeset))
      {
        if(!outpeerconnected)
        {
          outpeerconnected = 1;

          memcpy(outbufferop, handshake, sizeof(handshake));
          outbufferop_state = sizeof(handshake);
        }

        if(outbufferop_state)
        {
          result = send(outpeerfd, outbufferop, outbufferop_state, 0);

          if(result <= 0)
          {
            close(outpeerfd);
            outpeerfd = INVALID_SOCKET;

            continue;
          }

          outbufferop_state -= result;
          memmove(outbuffer, outbufferop + result, outbufferop_state);
        }
      }

      if(inpeerfd != INVALID_SOCKET && outbufferip_state && FD_ISSET(inpeerfd, &writeset))
      {
        result = send(inpeerfd, outbufferip, outbufferip_state, 0);

        if(result <= 0)
        {
          close(inpeerfd);
          inpeerfd = INVALID_SOCKET;

          continue;
        }

        outbufferip_state -= result;
        memmove(outbuffer, outbufferip + result, outbufferip_state);
      }

      if(outpeerfd != INVALID_SOCKET && FD_ISSET(outpeerfd, &readset))
      {
        struct data_packet* packet = (struct data_packet*) inbufferop;

        result = recv(outpeerfd, inbufferop + inbufferop_state,
                      sizeof(inbufferop) - inbufferop_state, 0);

        if(result <= 0)
        {
          close(outpeerfd);
          outpeerfd = INVALID_SOCKET;

          continue;
        }

        inbufferop_state += result;

        if(!outpeervalid)
        {
          if(inbufferop_state >= 4)
          {
            if(memcmp(inbufferop, handshake, 4))
            {
              close(outpeerfd);
              outpeerfd = INVALID_SOCKET;

              continue;
            }
            else
            {
              outpeervalid = 1;
              inbufferop_state -= 4;
              memmove(inbufferop, inbufferop + 4, inbufferop_state);
            }
          }
        }

        if(outpeervalid)
        {
          while(inbufferop_state >= 3 && inbufferop_state >= 3 + ntohs(packet->payload_size))
          {
            packet->payload_size = ntohs(packet->payload_size);
            process_packet(packet);

            inbufferop_state -= 3 + packet->payload_size;

            memmove(inbufferop, &inbufferop[3 + packet->payload_size],
                    inbufferop_state);
          }
        }
      }

      if(inpeerfd != INVALID_SOCKET && FD_ISSET(inpeerfd, &readset))
      {
        struct data_packet* packet = (struct data_packet*) inbufferip;

        result = recv(inpeerfd, inbufferip + inbufferip_state,
                      sizeof(inbufferip) - inbufferip_state, 0);

        if(result <= 0)
        {
          close(inpeerfd);
          inpeerfd = INVALID_SOCKET;

          continue;
        }

        inbufferip_state += result;

        if(!inpeervalid)
        {
          if(inbufferip_state >= 4)
          {
            if(memcmp(inbufferip, handshake, 4))
            {
              close(inpeerfd);
              inpeerfd = INVALID_SOCKET;

              continue;
            }
            else
            {
              inpeervalid = 1;
              inbufferip_state -= 4;
              memmove(inbufferip, inbufferip + 4, inbufferip_state);
            }
          }
        }

        if(inpeervalid)
        {
          while(inbufferip_state >= 3 && inbufferip_state >= 3 + ntohs(packet->payload_size))
          {
            packet->payload_size = ntohs(packet->payload_size);
            process_packet(packet);

            inbufferip_state -= 3 + packet->payload_size;

            memmove(inbufferip, &inbufferip[3 + packet->payload_size],
                    inbufferip_state);
          }
        }
      }
    }

    if(mode == MODE_ABORT_MESSAGE || mode == MODE_WAITING)
    {
      Sleep(30);

      rect.x = 0;
      rect.y = 0;
      rect.w = 640;
      rect.h = 480;

      SDL_BlitSurface(logo, &rect, screen, &rect);

      print_string(320, 250, message);
    }
    else if(mode == MODE_SCORE)
    {
      wchar_t buf[64];

      Sleep(30);

      rect.x = 0;
      rect.y = 0;
      rect.w = 640;
      rect.h = 480;

      SDL_BlitSurface(logo, &rect, screen, &rect);

      if(won)
        print_string(320, 250, L"You Won this Round!");
      else
        print_string(320, 250, L"You Lost this Round!");

#ifndef WIN32
      swprintf(buf, sizeof(buf), L"Won %u of %u Rounds", players[0].score, players[0].score + players[1].score);
#else
      swprintf(buf, L"Won %u of %u Rounds", players[0].score, players[0].score + players[1].score);
#endif

      print_string(320, 290, buf);

      if(players[0].ready && !players[1].ready)
        print_string(320, 330, L"Waiting for Other Player");
      else
      {
        if(!players[0].ready && players[1].ready)
          print_string(320, 330, L"Other Player is Waiting");

        print_string(320, 370, L"Press Space");
      }
    }
    else if(mode == MODE_MAIN_MENU)
    {
      int bump = (int) fabsf(sinf(last_tick * 0.2f) * 10.0f) - 5;
      ++last_tick;

      Sleep(15);

      rect.x = 0;
      rect.y = 200;
      rect.w = 640;
      rect.h = 180;

      SDL_BlitSurface(logo, &rect, screen, &rect);

      print_string(320 + (selection == 0 ? bump : 0), 250,
                   L"Join Random Game");
      print_string(320 + (selection == 1 ? bump : 0), 290,
                   L"Start Named Game");
      print_string(320 + (selection == 2 ? bump : 0), 330,
                   L"Join Named Game");
      print_string(320 + (selection == 3 ? bump : 0), 370,
                   L"Quit");
    }
    else if(mode == MODE_CREATE_GAME || mode == MODE_JOIN_GAME)
    {
      Sleep(30);

      rect.x = 0;
      rect.y = 200;
      rect.w = 640;
      rect.h = 180;

      SDL_BlitSurface(logo, &rect, screen, &rect);

      print_string(320, 250, L"Enter Game Name:");
      print_string(320, 290, game_name);
    }
    else if(mode == MODE_GAME)
    {
      Uint32 now = SDL_GetTicks();

      Sleep(5);

      while(now >= last_tick + time_stepms)
      {
        /* Skip delays larger than 0.5s */
        if(now - last_tick > 500)
          last_tick = now - 500;

        last_tick += time_stepms;

        game_tick(now - last_tick < time_stepms);
      }
    }

    while(SDL_PollEvent(&event))
    {
      switch(event.type)
      {
      case SDL_QUIT:

        exit(EXIT_SUCCESS);

        break;

      case SDL_KEYDOWN:

        if(event.key.keysym.unicode >= 32
        && has_char(event.key.keysym.unicode)
        && (mode == MODE_CREATE_GAME || mode == MODE_JOIN_GAME))
        {
          int length = 0;

          while(length < 32 && game_name[length])
            ++length;

          if(length < 32)
          {
            game_name[length] = event.key.keysym.unicode;
            game_name[length + 1] = 0;
          }

          continue;
        }

        switch(event.key.keysym.sym)
        {
        case SDLK_ESCAPE:

          if(mode == MODE_MAIN_MENU)
          {
            show_splash();

            exit(EXIT_SUCCESS);
          }
          else
          {
            if(mode == MODE_WAITING || mode == MODE_SCORE || mode == MODE_GAME)
            {
              struct data_packet p;

              p.packet_type = packet_abort_game;
              p.payload_size = 0;

              send_packet(&p);
            }

            SDL_BlitSurface(logo, 0, screen, 0);

            mode = MODE_MAIN_MENU;
            single_player = 0;
          }

          break;

        case SDLK_RETURN:

          if(mode == MODE_CREATE_GAME)
          {
            int i;
            struct data_packet p;

            if(!game_name[0])
              break;

            p.packet_type = packet_create_game;
            p.payload_size = sizeof(p.create_game);

            for(i = 0; i < 32; ++i)
              p.create_game.name[i] = game_name[i];

            if(INVALID_SOCKET == serverfd)
              connect_to_master();

            if(INVALID_SOCKET == serverfd)
            {
              wcscpy(message, L"Failed to Connect to Master Server");

              mode = MODE_ABORT_MESSAGE;

              break;
            }

            send_packet(&p);

            wcscpy(message, L"Waiting for Other Player");

            mode = MODE_WAITING;

            break;
          }
          else if(mode == MODE_JOIN_GAME)
          {
            int i;
            struct data_packet p;

            if(!game_name[0])
              break;

            p.packet_type = packet_join_game;
            p.payload_size = sizeof(p.join_game);

            for(i = 0; i < 32; ++i)
              p.join_game.name[i] = game_name[i] & 0xFF;

            if(INVALID_SOCKET == serverfd)
              connect_to_master();

            if(INVALID_SOCKET == serverfd)
            {
              wcscpy(message, L"Failed to Connect to Master Server");

              mode = MODE_ABORT_MESSAGE;

              break;
            }

            send_packet(&p);

            wcscpy(message, L"Waiting for Other Player");

            mode = MODE_WAITING;

            break;
          }

        case SDLK_SPACE:

          if(mode == MODE_MAIN_MENU)
          {
            if(selection == 0)
            {
              struct data_packet packet;

              packet.packet_type = packet_join_random_game;
              packet.payload_size = 0;

              if(INVALID_SOCKET == serverfd)
                connect_to_master();

              if(INVALID_SOCKET == serverfd)
              {
                wcscpy(message, L"Failed to Connect to Master Server");

                mode = MODE_ABORT_MESSAGE;

                break;
              }

              send_packet(&packet);

              single_player = 1;

              rng_seed = time(0);

              SDL_BlitSurface(background, 0, screen, 0);

              init_field();
              players[0].score = 0;
              players[1].score = 0;
              players[0].ready = 1;
              players[1].ready = 1;

              last_tick = SDL_GetTicks();
              game_tick(1);

              mode = MODE_GAME;
            }
            else if(selection == 1)
            {
              single_player = 0;
              memset(game_name, 0, sizeof(game_name));
              mode = MODE_CREATE_GAME;
            }
            else if(selection == 2)
            {
              single_player = 0;
              memset(game_name, 0, sizeof(game_name));
              mode = MODE_JOIN_GAME;
            }
            else if(selection == 3)
            {
              show_splash();

              exit(EXIT_SUCCESS);
            }
          }
          else if(mode == MODE_GAME)
          {
            if(!has_shot && !input_locked)
            {
              update_movement = 1;
              has_shot = 1;
              shoot(&players[0]);
            }
          }
          else if(mode == MODE_ABORT_MESSAGE)
          {
            mode = MODE_MAIN_MENU;
          }
          else if(mode == MODE_SCORE)
          {
            if(!players[0].ready)
            {
              struct data_packet p;

              players[0].ready = 1;

              init_field();

              init_movement_packet(&p);

              p.movement.flags |= FLAG_READY;

              send_packet(&p);

              if(players[1].ready)
              {
                SDL_BlitSurface(background, 0, screen, 0);

                mode = MODE_GAME;

                last_tick = SDL_GetTicks();
                game_tick(1);
              }
            }
          }

          break;

        case SDLK_UP:

          if(mode == MODE_MAIN_MENU)
          {
            last_tick = 0;

            if(selection == 0)
              selection = 3;
            else
              --selection;
          }
          else if(mode == MODE_GAME)
          {
            if(!has_shot && !input_locked)
            {
              update_movement = 1;
              has_shot = 1;
              shoot(&players[0]);
            }
          }

          break;

        case SDLK_DOWN:

          if(mode == MODE_MAIN_MENU)
          {
            last_tick = SDL_GetTicks();

            if(selection == 3)
              selection = 0;
            else
              ++selection;
          }

          break;

        case SDLK_LEFT:

          if(mode == MODE_GAME)
          {
            update_movement = 1;
            players[0].right = -1;
          }

          break;

        case SDLK_RIGHT:

          if(mode == MODE_GAME)
          {
            update_movement = 1;
            players[0].right = 1;
          }

          break;

        case SDLK_BACKSPACE:

          if(mode == MODE_CREATE_GAME || mode == MODE_JOIN_GAME)
          {
            int length = 0;

            while(length < 32 && game_name[length])
              ++length;

            if(length > 0)
              game_name[--length] = 0;
          }

          break;

        case 'f':

#ifndef WIN32
          SDL_WM_ToggleFullScreen(screen);
#else
          if(fullscreen)
            screen = SDL_SetVideoMode(width, height, 0, SDL_SWSURFACE);
          else
            screen = SDL_SetVideoMode(width, height, 0, SDL_SWSURFACE | SDL_FULLSCREEN);

          if(mode != MODE_GAME)
          {
            SDL_BlitSurface(logo, 0, screen, 0);
          }
          else
          {
            SDL_BlitSurface(background, 0, screen, 0);

            for(i = 0; i < 2; ++i)
            {
              players[i].dirty_minx = 0;
              players[i].dirty_miny = 0;
              players[i].dirty_maxx = max_field_width * 32;
              players[i].dirty_maxy = 440;
            }
          }
#endif
          fullscreen = !fullscreen;

          break;

        default:;
        }

        break;

      case SDL_KEYUP:

        switch(event.key.keysym.sym)
        {
        case SDLK_LEFT:

          if(players[0].right == -1)
          {
            update_movement = 1;
            players[0].right = 0;
          }

          break;

        case SDLK_RIGHT:

          if(players[0].right == 1)
          {
            update_movement = 1;
            players[0].right = 0;
          }

          break;

        default:;
        }

        break;
      }
    }

    if(update_movement)
    {
      struct data_packet packet;

      init_movement_packet(&packet);

      if(has_shot)
      {
        packet.movement.flags |= FLAG_BUBBLES;

        k = 0;

        memset(packet.movement.field, 0, sizeof(packet.movement.field));

        for(i = 0; i < field_height; ++i)
        {
          for(j = 0; j < WIDTH(i); ++j, ++k)
          {
            packet.movement.field[k / 2]
              |= players[0].field[i][j] << ((k & 1) ? 4 : 0);
          }
        }

        packet.movement.mbubble_count = 0;
        packet.payload_size += 54;

        for(i = 0; i < sizeof(players[0].mbubbles) / sizeof(players[0].mbubbles[0]); ++i)
        {
          if(players[0].mbubbles[i].color != 0)
          {
            j = packet.movement.mbubble_count++;
            packet.payload_size += sizeof(struct network_mbubble);

            pack_float(packet.movement.mbubbles[j].x, players[0].mbubbles[i].x);
            pack_float(packet.movement.mbubbles[j].y, players[0].mbubbles[i].y);
            pack_float(packet.movement.mbubbles[j].velx, players[0].mbubbles[i].velx);
            pack_float(packet.movement.mbubbles[j].vely, players[0].mbubbles[i].vely);
            packet.movement.mbubbles[j].falling = players[0].mbubbles[i].falling;
            packet.movement.mbubbles[j].color = players[0].mbubbles[i].color;
          }
        }
      }

      send_packet(&packet);
    }

    SDL_UpdateRect(screen, 0, 0, 0, 0);
  }
}
